/*
 *	CritLoops by Paul Hsieh
 *
 *	C++ test.
 *
 *  Copyright (c) 2012 Paul Hsieh
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *      Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *
 *      Redistributions in binary form must reproduce the above copyright notice,
 *      this list of conditions and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 *      Neither the name of CritLoops nor the names of its contributors may be used
 *      to endorse or promote products derived from this software without
 *      specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#if defined (_MSC_VER)
#define _CRT_SECURE_NO_WARNINGS
#endif

#if defined(__TURBOC__) && !defined (__BORLANDC__)
#define NOCPP
#define LIMITED_MEMORY
#endif

#include <stdio.h>      // for printf
#include <stdlib.h>     // for malloc()
#include <string.h>     // for Clib str* functions
#include <time.h>       // for clock()
#include <math.h>       // for pow()
#include <limits.h>     // for LONG_MAX
#include "bstrlib.h"	// for C bstring functions.

#define TESTSTRING1 "CritLoop benchmark"
#define TESTSTRING2 "Dot. date. datum. 123. Some more doubtful demonstration dummy data."

#ifndef NOCPP
#include <string>       // for std::string
#include "bstrwrap.h"	// for Bstrlib::CBString

/*
 *  Basic string concatenation performance testing.  The amount
 *  of concatenation is exaggerated versus real world applications
 *  in order to simulate the most stressful corner cases.  Various
 *  string implementation strategies, including length delimiting,
 *  copy on write, and fully referenced based collections can have
 *  a huge impact on the performance of string manipulation.
 */

/* Using STL. (Comparable to serious HLLs) */
static int stringConcatenation_STL (long count) {
	int i, c = 0;
	long int j;
	std::string a (TESTSTRING1);
	std::string accum;

	for (j=0; j < count; j++) {
		accum = "";
		for (i=0; i < 250; i++) {
			accum += a;	// <-- Bottleneck here.
			accum += "!!";
		}
		c += (int) accum.size ();
	}
	return c;
}

/* Using "The Better String Library". (Comparable to serious HLLs) */
static int stringConcatenation_Bstrlib_Cpp (long count) {
	int i, c = 0;
	long int j;
	Bstrlib::CBString a (TESTSTRING1);
	Bstrlib::CBString accum;

	for (j=0; j < count; j++) {
		accum = "";
		for (i=0; i < 250; i++) {
			accum += a;	// <-- Bottleneck here.
			accum += "!!";
		}
		c += accum.length ();
	}
	return c;
}

static int stringSearch_STL (long count) {
int i, c = 0;
long j;
std::string b (TESTSTRING2);

		for (j=0; j < count; j++) {
			for (c=i=0; i < 250; i++) {
				c += (int) b.find ("ummy");
				c += (int) b.find (" data");
				c += (int) b.find_first_of ("by");
			}
		}
		return c;
}

static int stringSearch_Bstrlib_Cpp (long count) {
int i, c = 0;
long j;
Bstrlib::CBString b (TESTSTRING2);

	for (j=0; j < count; j++) {
		for (c=i=0; i < 250; i++) {
			c += b.find ("ummy");
			c += b.find (" data");
			c += b.findchr ("by");
		}
	}
	return c;
}

#endif

/* Use standard C library functions, and "traverse the buffer overflow
   landmines" approach. */
static int stringConcatenation_Clib (long count) {
	int i, c = 0;
	long int j;
	char * accum = (char *) malloc (count * 250 * (sizeof (TESTSTRING1)-1 + sizeof ("!!")-1) + 1);
	const char * a = TESTSTRING1;

	if (NULL == accum) return -__LINE__;
	for (j=0; j < count; j++) {
		accum[0] = '\0';
		for (i=0; i < 250; i++) {
			strcat (accum, a);		/* <-- Bottleneck here. */
			strcat (accum, "!!");
		}
		c += (int) strlen (accum);
	}
	free (accum);
	return c;
}

/* Using "The Better String Library". (Comparable to serious HLLs) */
static int stringConcatenation_Bstrlib_C (long count) {
	int i, c = 0;
	long int j;
	struct tagbstring a = bsStatic (TESTSTRING1);
	bstring accum = bfromcstr ("");

	for (j=0; j < count; j++) {
		accum->slen = 0;
		for (i=0; i < 250; i++) {
			bconcat (accum, &a); /* <-- Bottleneck here. */
			bcatcstr (accum, "!!");
		}
		c += accum->slen;
	}
	return c;
}

/*
 *  Basic string searching performance testing.  String searching
 *  is commonly the counter balance to the most aggressive string
 *  implementation strategies which emphasize concatenation speed.
 *  It is difficult to be completely optimal on both.  The scenarios
 *  run here also stress the algorithmic efficiency of string
 *  searching.
 */

static int stringSearch_Clib (long count) {
int i, c = 0;
long j;
char * p;
char b[sizeof(TESTSTRING2)];

	strcpy (b, TESTSTRING2);
	for (j=0; j < count; j++) {
		for (c=i=0; i < 250; i++) {
			if (NULL != (p = strstr (b, "ummy"))) c += (int) (p - b);
			if (NULL != (p = strstr (b, " data"))) c += (int) (p - b);
			c += (int) strcspn (b, "by");
		}
	}
	return c;
}

static int stringSearch_Bstrlib_C (long count) {
int i, c = 0;
long j;
struct tagbstring b = bsStatic (TESTSTRING2);
struct tagbstring f1 = bsStatic ("ummy");
struct tagbstring f2 = bsStatic (" data");
struct tagbstring f3 = bsStatic ("by");

	for (j=0; j < count; j++) {
		for (c=i=0; i < 250; i++) {
			c += binstr (&b, 0, &f1);
			c += binstr (&b, 0, &f2);
			c += binchr (&b, 0, &f3);
		}
	}
	return c;
}

/*
 *  This heap sort uses a not so common implementation that performs
 *  the major data swapping as a rotation operation.  This generally
 *  reduces data bandwidth by 50% over the common swap-based
 *  implementation.  It also peels off and simplifies the top-most
 *  case when walking down the heaps.  Finally, it renders the
 *  inner-most comparison operation in such a way that it can be
 *  implemented as a conditional computation if the compiler and
 *  platform allow it.
 */

/* Assure the heap property for entries from top to last */
#define heapSortSiftDown(type, array, cmp, top, last) { \
int hssd__top = (top); \
int hssd__last = (last); \
int hssd__maxIdx = hssd__top; \
type hssd__tmp = array[hssd__top]; \
	while (hssd__last >= (hssd__maxIdx += hssd__maxIdx)) { \
		if (hssd__maxIdx != hssd__last && ((cmp (array[hssd__maxIdx], array[hssd__maxIdx + 1])) < 0)) hssd__maxIdx++; \
		if (cmp (hssd__tmp, array[hssd__maxIdx]) >= 0) break; \
		array[hssd__top] = array[hssd__maxIdx]; \
		hssd__top = hssd__maxIdx; \
	} \
	array[hssd__top] = hssd__tmp; \
}

/* Peel off the top siftDown operation since its parameters are trivial to
   fill in directly (and this saves us some moves.) */
#define heapSortSiftDown0(type, array, cmp, last) { \
type hssd0__tmp; \
int hssd0__last = (last); \
	if (cmp (array[1], array[0]) > 0) { \
		hssd0__tmp = array[1]; \
		array[1] = array[0]; \
		heapSortSiftDown (type, array, cmp, 1, hssd0__last); \
	} else { \
		hssd0__tmp = array[0]; \
	} \
	array[0] = array[hssd0__last]; \
	array[hssd0__last] = hssd0__tmp; \
}

#define heapSort(type, array, cmp, qty) { \
int hs__qty = (qty); \
type * hs__arr = (array); \
	if (hs__qty > 1) { \
		int hs__i = hs__qty / 2; \
		for (      hs__qty--; hs__i > 0; hs__i--) heapSortSiftDown  (type, hs__arr, cmp, hs__i, hs__qty); \
		for (hs__i = hs__qty; hs__i > 0; hs__i--) heapSortSiftDown0 (type, hs__arr, cmp, hs__i); \
	} \
}

#define HSA_LEN 4099
#define HSA_SKIP 131

#define scalarCmp(a,b) ((a)-(b))

static int heapSortInt (long count) {
	static int array[HSA_LEN];
	int i, j;
	long k;
	for (i=j=0; i < HSA_LEN; i++, j += HSA_SKIP) {
		if (j >= HSA_LEN) j -= HSA_LEN;
		array[i] = j;
	}
	for (k=0; k < count; k++) {
		heapSort (int, array, scalarCmp, HSA_LEN);
	}
	return array[HSA_LEN-1];
}

static int heapSortFp (long count) {
	static double array[HSA_LEN];
	int i, j;
	long k;
	for (i=j=0; i < HSA_LEN; i++, j+=HSA_SKIP) {
		if (j >= HSA_LEN) j -= HSA_LEN;
		array[i] = (double) j;
	}
	for (k=0; k < count; k++) {
		heapSort (double, array, scalarCmp, HSA_LEN);
	}
	return (int) array[HSA_LEN-1];
}

/*
 *  Based on the "Collatz Sequence" open math problem.  The function
 *  computes how many steps are required before the terminal sequence
 *  4, 2, 1, 4 is encountered (the open conjecture is that this
 *  sequence is always reached from any starting condition.)
 */

#define ME ((LONG_MAX-1)/3)
/* #define COLLATZ_SMALL_TABLE_QTY 16384 */

#define COLLATZ_SMALL_TABLE_QTY 8192

static int collatz (long count) {
long i, j, c = 0;
int mv;
static short iter[COLLATZ_SMALL_TABLE_QTY];

	for (j=0; j < count; j++) {
		long mx;

		for (mx=0; mx < COLLATZ_SMALL_TABLE_QTY; mx++) {
			iter[mx] = -1;
		}

		mv = 0;
		mx = ~0;

		for (i=1; i < 113383; i++) {
			long v = i;
			int curr = 0;

			while (v > 1) {
				if (v > ME && (v & 1)) { /* Predictable. */
					printf ("Collatz Overflow: %ld\n", i);
					curr = -1;
					break;
				}

				if (v < COLLATZ_SMALL_TABLE_QTY) { /* Non-predicatable but clustered. */
					if (iter[v] >= 0) { /* Asymptotically predictable. */
						curr += iter[v];
						break;
					}
				}

				if (v & 1) { /* Unpredictable. */
					v = (3*v + 1) >> 1;
				} else {
					v >>= 1;
				}

				curr ++;
			}

			if (curr > mv) { /* Highly biased. */
				mv  = curr;
				mx = i;
			}

			if (i < COLLATZ_SMALL_TABLE_QTY && curr < SHRT_MAX) {
				iter[i] = (short) curr;
			}
		}

		if (mx != 106239l || mv != 223) {
			printf ("Unexpected result: |%ld| -> %d\n", mx, mv);
		}

		c += mv;
	}

	return c;
}


/*
 *  These are the common "Stream" benchmarks for testing memory
 *  bandwidth for floating point operations.  Basically it measures
 *  the ability to sustain 1 or 2 streams read from memory and 1
 *  stream written while performing very light data dependent
 *  floating point operations.  Compilers have the opportunity to
 *  detect that the arrays are non-aliasing and rerender the loops
 *  into blocks for platforms which benefit from such
 *  transformations.  Alternatively, platforms/CPUs which can
 *  handle multiple streams of simultaneous data access can remove
 *  the need for better compilers.
 *
 *  Note: these arrays must be constructed with a parameter that
 *  guarantees that the arrays will not fit into any processor or
 *  system cache.  The base size is 2097152.  The time results
 *  should be scalled accordingly.
 */

/* Large enough, in 8 byte units, to exceed all caches in the system */

#define DEFAULT_STREAM_SZ (2l*1024l*1024l)

#if defined(LIMITED_MEMORY)
#define STREAM_SZ 2048
#else
#define STREAM_SZ DEFAULT_STREAM_SZ
#endif

#define LLTEST_SZ 2l*STREAM_SZ

static double * stream_A = NULL;
static double * stream_B = NULL;
static double * stream_C = NULL;
unsigned long * lltest = NULL;

static int streamInit (void) {
	long i;
	stream_A = (double *) malloc (STREAM_SZ * sizeof (double));
	stream_B = (double *) malloc (STREAM_SZ * sizeof (double));
	stream_C = (double *) malloc (STREAM_SZ * sizeof (double));
	lltest   = (unsigned long *) malloc (LLTEST_SZ * sizeof (unsigned long));
	if (NULL == stream_C || NULL == stream_B || NULL == stream_A || NULL == lltest) {
		free (stream_A);
		free (stream_B);
		free (stream_C);
		free (lltest);
		return -__LINE__;
	}
	for (i=0; i < STREAM_SZ; i++) {
		stream_A[i] = 1.0;
		stream_B[i] = 3.0;
	}
	for (i=0; i < LLTEST_SZ; i++) {
		unsigned long f = i + 43711ul;
		while (f >= LLTEST_SZ) f -= LLTEST_SZ;
		lltest[i] = f;
	}
	return 0;
}

static void streamUninit (void) {
	free (stream_A); stream_A = NULL;
	free (stream_B); stream_B = NULL;
	free (stream_C); stream_C = NULL;
	free (lltest); lltest = NULL;
}

#if !defined(LIMITED_MEMORY)
static int streamCopy (long count) {
	long i, j;
	if (NULL == stream_A) return -__LINE__;
	for (j=0; j < count; j++) {
		for (i=0; i < STREAM_SZ; i++) stream_C[i] = stream_A[i];
	}
	return 0;
}

static int streamScale (long count) {
	long i, j;
	if (NULL == stream_A) return -__LINE__;
	for (j=0; j < count; j++) {
		for (i=0; i < STREAM_SZ; i++) stream_C[i] = 3.0*stream_A[i];
	}
	return 0;
}

static int streamAdd (long count) {
	long i, j;
	if (NULL == stream_A) return -__LINE__;
	for (j=0; j < count; j++) {
		for (i=0; i < STREAM_SZ; i++) stream_C[i] = stream_A[i] + stream_B[i];
	}
	return 0;
}

static int streamTriad (long count) {
	long i, j;
	if (NULL == stream_A) return -__LINE__;
	for (j=0; j < count; j++) {
		for (i=0; i < STREAM_SZ; i++) stream_C[i] = stream_A[i] + 3.0*stream_B[i];
	}
	return 0;
}
#endif

/* Perform a linked list traversal test in a large section of memory. */
static int latencyRAM (long count) {
	long i, j;
	unsigned long k;
	k = 0;
	for (j=0; j < count; j++) {
		for (i=0; i < 1024; i++) {
			k = lltest[k];
		}
	}
	return (int) k;
}

int kinit = 0;

/* Perform a linked list traversal test in a very small section of memory. */
static int latencyL1Cache (long count) {
	long i, j;
	unsigned long k = kinit;

	for (i=0; i < 1024; i++) {
		lltest[i] &= 1023;
	}
	for (j=0; j < count; j++) {
		for (i=0; i < 1024/8; i++) {
			k = lltest[k];
			k = lltest[k];
			k = lltest[k];
			k = lltest[k];
			k = lltest[k];
			k = lltest[k];
			k = lltest[k];
			k = lltest[k];
		}
	}
	return (int) k;
}

/*
 *  A standard 4x4 matrix multiplication.  Although matrix by vector
 *  is more common, it can be considered a subset of this operation
 *  (the performance characteristics should be the same).  One matrix
 *  is assumed (runtime) constant and pre-transposed for better memory
 *  access.  The most inner loop is unrolled, as it is assumed that
 *  developers interested in this kernel would be concerned with
 *  performance from the start.  The primary platform potential here
 *  should be to leverage the SIMD capabilities of modern processors.
 */

/* This needs to be extern otherwise the compiler may over-simplify */

float preTransposed[4][4] = {
	{ 0, 1, 0, 0 },
	{ 1, 0, 0, 0 },
	{ 0, 0, 1, 0 },
	{ 0, 0, 0, 1 }
};

static int MatrixMulGfx (long count) {
	float source[4][4];
	float target[4][4];
	int i, j;
	long k;

	for (i=0; i < 4; i++) {
		for (j=0; j < 4; j++) {
			source[i][j] = rand() / (float) (RAND_MAX+1.0);
			target[i][j] = 0;
		}
	}

	for (k=0; k < count; k++) {
		for (j=0; j < 1000; j++) {
			for (i=0; i < 4; i++) {
				/* Unrolled -- as opposed to FPUMark which was a loop */
				target[i][0] = preTransposed[0][0] * source[i][0]
							 + preTransposed[0][1] * source[i][1]
							 + preTransposed[0][2] * source[i][2]
							 + preTransposed[0][3] * source[i][3];
				target[i][1] = preTransposed[1][0] * source[i][0]
							 + preTransposed[1][1] * source[i][1]
							 + preTransposed[1][2] * source[i][2]
							 + preTransposed[1][3] * source[i][3];
				target[i][2] = preTransposed[2][0] * source[i][0]
							 + preTransposed[2][1] * source[i][1]
							 + preTransposed[2][2] * source[i][2]
							 + preTransposed[2][3] * source[i][3];
				target[i][3] = preTransposed[3][0] * source[i][0]
							 + preTransposed[3][1] * source[i][1]
							 + preTransposed[3][2] * source[i][2]
							 + preTransposed[3][3] * source[i][3];
			}
			for (i=0; i < 4; i++) {
				source[i][0] = target[i][0];
				source[i][1] = target[i][1];
				source[i][2] = target[i][2];
				source[i][3] = target[i][3];
			}
		}
	}
	return (int) (100.0 * target[3][3]);
}

#define MAX_ITERS 2048

static int mandelPointTest (double cr, double ci) {
double zr, zi;
int i;

	zr = cr; zi = ci;

	for (i=0; i < MAX_ITERS && zr*zr + zi*zi < 4.0; i++) {
		double tmp = zr * zr - zi * zi;
		zi = 2 * zr * zi + ci;
		zr = tmp + cr;
	}
	return i;
}

static int Mandelbrot (long count) {
	double cr = -2.0, ci = 2.0;
	int i, j, c;

	c = 0;
	for (j=0; j < count; j++) {
		for (i=0; i < 64; i++) {
			cr = (4.0 * i / 64.0) - 2.0;
			ci = 2.0 - (4.0 * i / 64.0);
			c += mandelPointTest (cr, ci);
		}
	}
	return c;
}

#ifndef CLOCKS_PER_SEC
# ifdef CLK_TCK
#  define CLOCKS_PER_SEC (CLK_TCK)
# else
#  error CLOCKS_PER_SEC not defined
# endif
#endif

#define DURATION 3.0

double test (int (* tfn) (long count)) {
	long int count;
	int ret;
	clock_t t0, t1, t2, t3;

	/*
	 *  Calibrate the test so that it runs for a projected
	 *  time of DURATION seconds.  This calibration step can
	 *  take between 0.4s and 0.8s to complete.
	 */
	count = 1;
	L1:;
	t0 = clock ();
	ret = tfn (count);
	t1 = clock ();
	if ((t1 - t0) < 0.2 * CLOCKS_PER_SEC) {
		count += count;
		if (count > 536870912l) printf ("! <%ld>, ret = %d\n", count, ret);
		if (0 >= count) {
			fprintf (stderr, "Test is way too fast\n");
			return -__LINE__;
		}
		goto L1;
	}
	/* Estimation for the number of counts per second */
	count = (long) (count * (double) CLOCKS_PER_SEC * DURATION) / (t1 - t0);
	if (0 >= count) count = 1;

	/* Run the test 3 times in a row. */
	t0 = clock ();
	ret += tfn (count);
	t1 = clock ();
	ret += tfn (count);
	t2 = clock ();
	ret += tfn (count);
	t3 = clock ();

	if (0 >= count + count) printf ("! <%ld>, ret = %d\n", count, ret);

	/* The time taken for each test in clock ticks. */
	t3 -= t2;
	t2 -= t1;
	t1 -= t0;

	/* Determine median of t1, t2, t3 (throw out highest and lowest) */
	if (t3 > t2) {
		t0 = t3;
		t3 = t2;
		t2 = t0;
	}
	if (t1 >= t2) t3 = t2;
	else if (t1 >= t3) t3 = t1;

	/* The iterations per second rate of the median test. */
	return (((double) CLOCKS_PER_SEC) * count) / t3;
}

#define CNAMEVER_aux(NAME,VER) NAME #VER
#define CNAMEVER(NAME,VER) CNAMEVER_aux(NAME,VER)

/*
 *  Vendors are encouraged to fill in the following macro with their
 *  compiler details.  Please also make a note if there are ordering
 *  requirements.
 */

/*
 *  Ordering notes:
 *   1) Intel must be detected before Microsoft and GNU.
 */

#if defined(__WATCOMC__)
#define COMPILERNAME CNAMEVER ("Watcom ", __WATCOMC__)
#elif defined(__MWERKS__)
#define COMPILERNAME CNAMEVER ("Metrowerks ", __MWERKS__)
#elif defined (__INTEL_COMPILER)
# define COMPILERNAME CNAMEVER ("Intel ", __INTEL_COMPILER)
#elif defined (__ICL)
# define COMPILERNAME CNAMEVER ("Intel ", __ICL)
#elif defined (__EDG__)
# define COMPILERNAME "Intel"
#elif defined(_MSC_VER)
#define COMPILERNAME CNAMEVER( "Microsoft ", _MSC_VER)
#elif defined(__BORLANDC__)
#define COMPILERNAME CNAMEVER ("Borland ", __BORLANDC__)
#elif defined(__TURBOC__)
#define COMPILERNAME CNAMEVER ("Turbo ", __TURBOC__)
#elif defined(__GNUC__)
# if defined(__GNUC_PATCHLEVEL__)
#  define __GNUC_VERSION__ __GNUC__ __GNUC_MINOR__ __GNUC_PATCHLEVEL__
# else
#  define __GNUC_VERSION__ __GNUC__ __GNUC_MINOR__
# endif
#if defined (__CYGWIN__)
#define COMPILERNAME CNAMEVER ("Gnu Cygwin ", __GNUC_VERSION__)
#elif defined(__DJGPP__)
#define COMPILERNAME CNAMEVER ("Gnu DJGPP ", __GNUC_VERSION__)
#elif defined(__MINGW32__)
#define COMPILERNAME CNAMEVER ("Gnu MinGW ", __GNUC_VERSION__)
#else
#define COMPILERNAME CNAMEVER ("Gnu ", __GNUC_VERSION__)
#endif
#elif defined(__DJGPP__)
#define __DJGPP_VERSION__ __DJGPP__ __DJGPP_MINOR__
#define COMPILERNAME CNAMEVER ("DJGPP ", __DJGPP_VERSION__)
#elif defined(_CRAYC)
#define __CRAYC_VERSION__ _RELEASE _RELEASE_MINOR
#define COMPILERNAME CNAMEVER ("Cray ", __CRAYC_VERSION__)
#elif defined(__DCC__)
#define COMPILERNAME CNAMEVER ("Diab ", __VERSION_NUMBER__)
#elif defined(__PGI)
#define COMPILERNAME "Portland Group "
#else
#define COMPILERNAME "Unknown compiler vendor"
#endif

#define GREPPABLETAG "@CL@"

#define TEST(r,N) r = test(N); printf ("\tentry['C " COMPILERNAME "']['sub']['" #N "'] = %.3f; // " GREPPABLETAG "\n\tentry['C++ " COMPILERNAME "']['sub']['" #N "'] = %.3f; // " GREPPABLETAG "\n", r, r);

#if defined(NOCPP)
#define TESTCPP(r,N)
#else
#define TESTCPP(r,N) r = test(N); printf ("\tentry['C++ " COMPILERNAME "']['sub']['" #N "'] = %.3f; // " GREPPABLETAG "\n", r);
#endif

int main (int argc, char * argv[]) {
#if !defined(NOCPP)
	double strmcpp;
	double strmcpp_alt;
	double strscpp;
	double strscpp_alt;
	double strrate_cpp;
	double opt_strrate_cpp;
#endif

	double strmc;
	double strmc_alt;
	double strsc;
	double strsc_alt;
	double fpuband;
	double fpulat;
	double flow1;
	double flow2;
	double flow3;
#if !defined(LIMITED_MEMORY)
	double memband1;
	double memband2;
	double memband3;
	double memband4;
#endif
	double memlat1;
	double memlat2;

	double strrate_c;
	double fpurate;
	double flowrate;
#if !defined(LIMITED_MEMORY)
	double memrate;
#endif

	double opt_strrate_c;
	double opt_fpurate;
	double opt_flowrate;
#if !defined(LIMITED_MEMORY)
	double opt_memrate;
#endif

	argc = argc;
	argv = argv;
	printf ("CritLoops for C++ (" COMPILERNAME ")\n");

	printf ("\tentry['C++ " COMPILERNAME "']['sub'] = new Array(); // " GREPPABLETAG "\n");
	printf ("\tentry['C " COMPILERNAME "']['sub'] = new Array(); // " GREPPABLETAG "\n");

	TEST    (fpuband,     MatrixMulGfx);
	TEST    (fpulat,      Mandelbrot);
	TESTCPP (strmcpp,     stringConcatenation_STL);
	TEST    (strmc,       stringConcatenation_Clib);
	TESTCPP (strmcpp_alt, stringConcatenation_Bstrlib_Cpp);
	TEST    (strmc_alt,   stringConcatenation_Bstrlib_C);
	TESTCPP (strscpp,     stringSearch_STL);
	TEST    (strsc,       stringSearch_Clib);
	TESTCPP (strscpp_alt, stringSearch_Bstrlib_Cpp);
	TEST    (strsc_alt,   stringSearch_Bstrlib_C);
	TEST    (flow1,       heapSortInt);
	TEST    (flow2,       heapSortFp);
	TEST    (flow3,       collatz);
	streamInit ();
#if !defined(LIMITED_MEMORY)
	TEST    (memband1,    streamCopy);
	TEST    (memband2,    streamScale);
	TEST    (memband3,    streamAdd);
	TEST    (memband4,    streamTriad);
#endif
	TEST    (memlat1,     latencyRAM);
	TEST    (memlat2,     latencyL1Cache);
	streamUninit ();

	printf ("\tentry['C++ " COMPILERNAME "']['std-rate'] = new Array(); // " GREPPABLETAG "\n");
	printf ("\tentry['C " COMPILERNAME "']['std-rate'] = new Array(); // " GREPPABLETAG "\n");
	printf ("\tentry['C++ " COMPILERNAME "']['opt-rate'] = new Array(); // " GREPPABLETAG "\n");
	printf ("\tentry['C " COMPILERNAME "']['opt-rate'] = new Array(); // " GREPPABLETAG "\n");

	/* Compute standard results */

	fpurate     = pow (fpuband * fpulat, 0.5);
	strrate_c   = pow (strmc * strsc, 0.5);
#if !defined(NOCPP)
	strrate_cpp = pow (strmcpp * strscpp, 0.5);
#endif
	flowrate    = pow (flow1 * flow2 * flow3, 1/3.0);
#if !defined(LIMITED_MEMORY)
	memrate     = pow (memband1 * memband2 * memband3, 1/12.0)
	            * pow (memband4 * memlat1 * memlat2, 1/4.0);
#endif

	/* Compute peak results */

	/* Can use alternate functions */

	if (strmc_alt > strmc) strmc = strmc_alt;
	if (strsc_alt > strsc) strsc = strsc_alt;
#if !defined(NOCPP)
	if (strmcpp_alt > strmcpp) strmcpp = strmcpp_alt;
	if (strscpp_alt > strscpp) strscpp = strscpp_alt;

	/* Can use C in place of C++ */

	if (strmc > strmcpp) strmcpp = strmc;
	if (strsc > strscpp) strscpp = strsc;
#endif

	opt_fpurate     = fpurate;
	opt_strrate_c   = pow (strmc * strsc, 0.5);
#if !defined(NOCPP)
	opt_strrate_cpp = pow (strmcpp * strscpp, 0.5);
#endif
	opt_flowrate    = flowrate;
#if !defined(LIMITED_MEMORY)
	opt_memrate     = memrate;
#endif

	/* Give the results */

#if !defined(NOCPP)
	printf ("\tentry['C++ " COMPILERNAME "']['std-rate'][ 'Fpu'] = %.3f; // " GREPPABLETAG "\n", fpurate);
	printf ("\tentry['C++ " COMPILERNAME "']['std-rate'][ 'Str'] = %.3f; // " GREPPABLETAG "\n", strrate_cpp);
	printf ("\tentry['C++ " COMPILERNAME "']['std-rate']['Flow'] = %.3f; // " GREPPABLETAG "\n", flowrate);
	printf ("\tentry['C++ " COMPILERNAME "']['std-rate'][ 'Mem'] = %.3f; // " GREPPABLETAG "\n", memrate);
	printf ("\tentry['C++ " COMPILERNAME "']['std'] = %.3f; // " GREPPABLETAG "\n", pow (fpurate * strrate_cpp * flowrate * memrate, 1/4.0));
	printf ("\tentry['C++ " COMPILERNAME "']['opt-rate'][ 'Fpu'] = %.3f; // " GREPPABLETAG "\n", opt_fpurate);
	printf ("\tentry['C++ " COMPILERNAME "']['opt-rate'][ 'Str'] = %.3f; // " GREPPABLETAG "\n", opt_strrate_cpp);
	printf ("\tentry['C++ " COMPILERNAME "']['opt-rate']['Flow'] = %.3f; // " GREPPABLETAG "\n", opt_flowrate);
	printf ("\tentry['C++ " COMPILERNAME "']['opt-rate'][ 'Mem'] = %.3f; // " GREPPABLETAG "\n", opt_memrate);
	printf ("\tentry['C++ " COMPILERNAME "']['opt'] = %.3f; // " GREPPABLETAG "\n", pow (opt_fpurate * opt_strrate_cpp * opt_flowrate * opt_memrate, 1/4.0));
#endif

	printf ("\tentry['C " COMPILERNAME "']['std-rate'][ 'Fpu'] = %.3f; // " GREPPABLETAG "\n", fpurate);
	printf ("\tentry['C " COMPILERNAME "']['std-rate'][ 'Str'] = %.3f; // " GREPPABLETAG "\n", strrate_c);
	printf ("\tentry['C " COMPILERNAME "']['std-rate']['Flow'] = %.3f; // " GREPPABLETAG "\n", flowrate);
#if !defined(LIMITED_MEMORY)
	printf ("\tentry['C " COMPILERNAME "']['std-rate'][ 'Mem'] = %.3f; // " GREPPABLETAG "\n", memrate);
	printf ("\tentry['C " COMPILERNAME "']['std'] = %.3f; // " GREPPABLETAG "\n", pow (fpurate * strrate_c * flowrate * memrate, 1/4.0));
#endif
	printf ("\tentry['C " COMPILERNAME "']['opt-rate'][ 'Fpu'] = %.3f; // " GREPPABLETAG "\n", opt_fpurate);
	printf ("\tentry['C " COMPILERNAME "']['opt-rate'][ 'Str'] = %.3f; // " GREPPABLETAG "\n", opt_strrate_c);
	printf ("\tentry['C " COMPILERNAME "']['opt-rate']['Flow'] = %.3f; // " GREPPABLETAG "\n", opt_flowrate);
#if !defined(LIMITED_MEMORY)
	printf ("\tentry['C " COMPILERNAME "']['opt-rate'][ 'Mem'] = %.3f; // " GREPPABLETAG "\n", opt_memrate);
	printf ("\tentry['C " COMPILERNAME "']['opt'] = %.3f; // " GREPPABLETAG "\n", pow (opt_fpurate * opt_strrate_c * opt_flowrate * opt_memrate, 1/4.0));
#endif
	printf ("\n");

	return 0;
}
