/*
 *  CritLoops by Paul Hsieh
 *
 *  C# tests.
 *
 *  Copyright (c) 2012 Paul Hsieh
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *      Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *
 *      Redistributions in binary form must reproduce the above copyright notice,
 *      this list of conditions and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 *      Neither the name of CritLoops nor the names of its contributors may be used
 *      to endorse or promote products derived from this software without
 *      specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *  
 */

using System;
using System.Text;

namespace Bench {
	/// <summary>
	/// The C# implementation of the Open Bench benchmark.
	/// </summary>
	class Benchmark {
		/// <summary>
		/// Entry point.
		/// </summary>
        static String pname = "C#";

        [STAThread]
		static void Main(string[] args) {

			double strm;
			double strs;
			double fpuband;
			double fpubandalt;
			double fpulat;
			double flow1;
			double flow2;
			double flow3;
			double memband1;
			double memband2;
			double memband3;
			double memband4;
			double memlat1;
			double memlat2;

			double strrate;
			double fpurate;
			double flowrate;
			double memrate;

			double peak_strrate;
			double peak_fpurate;
			double peak_flowrate;
			double peak_memrate;

            Console.WriteLine("CritLoops for C#");
            
			if (args.Length > 0) {
                pname = args[0];
			}

            Console.WriteLine("\tentry['" + pname + "']['sub'] = new Array (); // @CL@");

			fpuband    = runtest (new matrixMulGfx ());
			fpubandalt = runtest (new matrixMulGfxCheat ());
			fpulat     = runtest (new mandelbrot ());
			strm       = runtest (new stringConcatTest ());
			strs       = runtest (new stringSearchTest ());
			flow1      = runtest (new heapSortIntTest ());
			flow2      = runtest (new heapSortFpTest ());
			flow3      = runtest (new collatzTest ());
			memband1   = runtest (new streamCopy ());
			memband2   = runtest (new streamScale ());
			memband3   = runtest (new streamAdd ());
			memband4   = runtest (new streamTriad ());
			memlat1    = runtest (new latencyRAM ());
			memlat2    = runtest (new latencyL1Cache ());

			/* Compute standard results */

			fpurate  = Math.Pow (fpuband * fpulat, 0.5);
			strrate  = Math.Pow (strm * strs, 0.5);
			flowrate = Math.Pow (flow1 * flow2 * flow3, 1/3.0);
			memrate  = Math.Pow (memband1 * memband2 * memband3, 1/12.0)
			         * Math.Pow (memband4 * memlat1 * memlat2, 1/4.0);

			/* Compute peak results */

			if (fpubandalt > fpuband) fpuband = fpubandalt;
			peak_fpurate  = Math.Pow (fpuband * fpulat, 0.5);
			peak_strrate  = strrate;
			peak_flowrate = flowrate;
			peak_memrate  = memrate;

			/* Give the results */

            Console.WriteLine("\tentry['" + pname + "']['std-rate'] = new Array(); // @CL@");
            Console.WriteLine("\tentry['" + pname + "']['opt-rate'] = new Array(); // @CL@");

            Console.WriteLine("\tentry['" + pname + "']['std-rate'][ 'Fpu'] = " + fpurate + "; // @CL@");
            Console.WriteLine("\tentry['" + pname + "']['std-rate'][ 'Str'] = " + strrate + "; // @CL@");
            Console.WriteLine("\tentry['" + pname + "']['std-rate']['Flow'] = " + flowrate + "; // @CL@");
            Console.WriteLine("\tentry['" + pname + "']['std-rate'][ 'Mem'] = " + memrate + "; // @CL@");
            Console.WriteLine("\tentry['" + pname + "']['std'] = " + Math.Pow (fpurate * strrate * flowrate * memrate, 1 / 4.0) + "; // @CL@");

            Console.WriteLine("\tentry['" + pname + "']['opt-rate'][ 'Fpu'] = " + peak_fpurate + "; // @CL@");
            Console.WriteLine("\tentry['" + pname + "']['opt-rate'][ 'Str'] = " + peak_strrate + "; // @CL@");
            Console.WriteLine("\tentry['" + pname + "']['opt-rate']['Flow'] = " + peak_flowrate + "; // @CL@");
            Console.WriteLine("\tentry['" + pname + "']['opt-rate'][ 'Mem'] = " + peak_memrate + "; // @CL@");
            Console.WriteLine("\tentry['" + pname + "']['opt'] = " + Math.Pow (peak_fpurate * peak_strrate * peak_flowrate * peak_memrate, 1 / 4.0) + "; // @CL@");

			Console.WriteLine ("");
		}

		public static double runtest (testCase tc) {
			long startTime = 1000;
			long stopTime = 0;
			long t0, t1, t2, t3;
			double ips, TICKS_PER_SECOND = 10000000.0;
			double DURATION = 3.0;
			int result;

			/* 
			 * Calibrate the test so that it runs for a projected 
			 * time of DURATION seconds.  This calibration step can
			 * take between 0.4s and 0.8s to complete.
			 */
			int count = 1;
			while (count > 0) {
				startTime = DateTime.Now.Ticks;
				result = tc.subTest (count);
				stopTime = DateTime.Now.Ticks;
				if (stopTime - startTime > 0.2 * TICKS_PER_SECOND) 
					break;
				count = count + count;
			}
			if (0 >= count) {
				Console.WriteLine (tc.name() + " is way too fast");
				return -1;
			}
			count = (int) (TICKS_PER_SECOND * (double) count * DURATION / (stopTime - startTime));
			if (0 >= count) count = 1;

			/*
			 *  Run the test 3 times.
			 */
			t0 = DateTime.Now.Ticks;
			result = tc.subTest (count);
			t1 = DateTime.Now.Ticks;
			result = tc.subTest (count);
			t2 = DateTime.Now.Ticks;
			result = tc.subTest (count);
			t3 = DateTime.Now.Ticks;

			/* The time taken for each test in clock ticks. */
			t3 -= t2;
			t2 -= t1;
			t1 -= t0;

			/* Determine median of t1, t2, t3 (throw out highest and lowest) */
			if (t3 > t2) {
				t0 = t3;
				t3 = t2;
				t2 = t0;
			}
			if (t1 >= t2) t3 = t2;
			else if (t1 >= t3) t3 = t1;

			/* 
			 *  The official score of the test in units of iterations 
			 *  per second.
			 */
			ips = (TICKS_PER_SECOND / t3) * count;
            Console.WriteLine ("\tentry['" + pname + "']['sub']['" + tc.name() + "'] = " + ips + "; // @CL@");

			/*
			 *  C# note: Try to minimize the impact of the garbage from
			 *  a previous test from interfering with the results of
			 *  subsequent tests.
			 */
			System.GC.Collect();
			return ips;
		}
	}

	/* Base class for all the sub-tests */
	abstract class testCase {
		/*
		 *  The subTest method is supposed to implement the core 
		 *  kernel that is to be tested for performance.  The count
		 *  parameter is the number of times the kernel should be 
		 *  run.  Each iteration of the run should be as independent
		 *  as possible from previous iterations.
		 */
		public abstract int subTest (int count);

		public abstract String name ();
	}

	/* Base class for the string tests */
	abstract class stringTest : testCase {
        static public String TESTSTRING1 = "CritLoop benchmark";
		static public String TESTSTRING2 = "Dot. date. datum. 123. Some more doubtful demonstration dummy data.";
	}

	/*
	 *  Basic string concatenation performance testing.  The amount
	 *  of concatenation is exaggerated versus real world applications
	 *  in order to simulate the most stressful corner cases.  Various
	 *  string implementation strategies, including length delimiting, 
	 *  copy on write, and fully referenced based collections can have 
	 *  a huge impact on the performance of string manipulation.
	 */
	class stringConcatTest : stringTest {
		override public int subTest (int count) {
			// StringBuilder accum = new StringBuilder("");
			// StringBuilder empty = new StringBuilder("");
            String accum = "";
            String empty = "";
			String a;
			int i, j, c = 0;

			a = TESTSTRING1;

			for (j=0; j < count; j++) {
				accum = empty;
				for (i=0; i < 250; i++) {
					// accum.Append (a);      // <-- Bottleneck here.
					// accum.Append ("!!");
                    accum += a;
                    accum += "!!";
				}
				c += accum.Length;
			}
			return c;
		}

		override public String name () {
			return "String Concatenation";
		}
	}

	/*
	 *  Basic string searching performance testing.  String searching 
	 *  is commonly the counter balance to the most aggressive string
	 *  implementation strategies which emphasize concatenation speed.
	 *  It is difficult to be completely optimal on both.  The scenarios
	 *  run here also stress the algorithmic efficiency of string 
	 *  searching.
	 */
	class stringSearchTest : stringTest {
		override public int subTest (int count) {
			String b = TESTSTRING2;
			char [] s = { 'b', 'y' };
			int i, j, c = 0;

			for (j=0; j < count; j++) {
				for (i=0; i < 250; i++) {
					c += b.IndexOf ("ummy");
					c += b.IndexOf (" data");
					c += b.IndexOfAny (s);
				}
			}
			return c;
		}

		override public String name () {
			return "String Scanning";
		}
	}

	/*
	 *  A standard 4x4 matrix multiplication.  Although matrix by vector
	 *  is more common, it can be considered a subset of this operation 
	 *  (the performance characteristics should be the same).  One matrix 
	 *  is assumed (runtime) constant and pre-transposed for better memory 
	 *  access.  The most inner loop is unrolled, as it is assumed that 
	 *  developers interested in this kernel would be concerned with
	 *  performance from the start.  The primary platform potential here 
	 *  should be to leverage the SIMD capabilities of modern processors.
	 */
	class matrixMulGfx : testCase {
		override public int subTest (int count) {
			float[,] preTransposed = new float[4,4] {
				{ 0, 1, 0, 0 },
				{ 1, 0, 0, 0 },
				{ 0, 0, 1, 0 },
				{ 0, 0, 0, 1 }
			};
			float[,] source = new float[4,4];
			float[,] target = new float[4,4];
			int i, j;
			long k;
			Random prng = new Random ();

			for (i=0; i < 4; i++) {
				for (j=0; j < 4; j++) {
					source[i,j] = (float) (prng.Next () / 2147483648.0);
				}
			}

			for (k=0; k < count; k++) {
				for (j=0; j < 1000; j++) {
					for (i=0; i < 4; i++) {
						/* Unrolled -- as opposed to FPUMark which was a loop */
						target[i,0] = preTransposed[0,0] * source[i,0]
						            + preTransposed[0,1] * source[i,1]
						            + preTransposed[0,2] * source[i,2]
						            + preTransposed[0,3] * source[i,3];
						target[i,1] = preTransposed[1,0] * source[i,0]
						            + preTransposed[1,1] * source[i,1]
						            + preTransposed[1,2] * source[i,2]
						            + preTransposed[1,3] * source[i,3];
						target[i,2] = preTransposed[2,0] * source[i,0]
						            + preTransposed[2,1] * source[i,1]
						            + preTransposed[2,2] * source[i,2]
						            + preTransposed[2,3] * source[i,3];
						target[i,3] = preTransposed[3,0] * source[i,0]
						            + preTransposed[3,1] * source[i,1]
						            + preTransposed[3,2] * source[i,2]
						            + preTransposed[3,3] * source[i,3];
					}
					for (i=0; i < 4; i++) {
						source[i,0] = target[i,0];
						source[i,1] = target[i,1];
						source[i,2] = target[i,2];
						source[i,3] = target[i,3];
					}
				}
			}
			return (int) (100.0 * target[3,3]);
		}

		override public String name () {
			return "Matrix Multiply GFX";
		}
	}

	class matrixMulGfxCheat : testCase {
		/*
		 *  C# note: Amazingly, for this benchmark, the Visual Studio 
		 *  .NET 2003 compiler is very susceptable to cheesy 
		 *  1980s-style optimizations.  This includes manual hoisting, 
		 *  common sub-expression elimination, explicit index mapping 
		 *  from a 2D to a 1D array and completely obvious unrolling.  
		 *  The fact that this transformation makes such a huge 
		 *  difference is indicative that its not CLR which is to 
		 *  blame, but rather the compiler.
		 */
		override public int subTest (int count) {
			float[] preTransposed = new float[16] {
				0, 1, 0, 0, 
				1, 0, 0, 0, 
				0, 0, 1, 0, 
				0, 0, 0, 1 };
			float[] source = new float[16];
			float[] target = new float[16];
			int i, j;
			int i0;
			long k;
			Random prng = new Random ();

			for (i=0; i < 16; i++) {
				source[i] = (float) (prng.Next () / 2147483648.0);
			}

			for (k=0; k < count; k++) {
				for (j=0; j < 1000; j++) {
					for (i0=0; i0 < 16; i0 += 4) {
						/* Unrolled -- as opposed to FPUMark which was a loop */
						target[i0] = preTransposed[0*4+0] * source[i0]
							+ preTransposed[0*4+1] * source[i0+1]
							+ preTransposed[0*4+2] * source[i0+2]
							+ preTransposed[0*4+3] * source[i0+3];
						target[i0+1] = preTransposed[1*4+0] * source[i0]
							+ preTransposed[1*4+1] * source[i0+1]
							+ preTransposed[1*4+2] * source[i0+2]
							+ preTransposed[1*4+3] * source[i0+3];
						target[i0+2] = preTransposed[2*4+0] * source[i0]
							+ preTransposed[2*4+1] * source[i0+1]
							+ preTransposed[2*4+2] * source[i0+2]
							+ preTransposed[2*4+3] * source[i0+3];
						target[i0+3] = preTransposed[3*4+0] * source[i0]
							+ preTransposed[3*4+1] * source[i0+1]
							+ preTransposed[3*4+2] * source[i0+2]
							+ preTransposed[3*4+3] * source[i0+3];
					}
					
					source[ 0] = target[ 0];
					source[ 1] = target[ 1];
					source[ 2] = target[ 2];
					source[ 3] = target[ 3];
					source[ 4] = target[ 4];
					source[ 5] = target[ 5];
					source[ 6] = target[ 6];
					source[ 7] = target[ 7];
					source[ 8] = target[ 8];
					source[ 9] = target[ 9];
					source[10] = target[10];
					source[11] = target[11];
					source[12] = target[12];
					source[13] = target[13];
					source[14] = target[14];
					source[15] = target[15];
				}
			}
			return (int) (100.0 * target[15]);
		}

		override public String name () {
			return "Matrix Multiply GFX Hack";
		}
	}

	class mandelbrot : testCase {
		 int MAX_ITERS = 2048;

		public int mandelPointTest (double cr, double ci) {
			double zr, zi;
			int i;

			zr = cr; zi = ci;

			for (i=0; i < MAX_ITERS && zr*zr + zi*zi < 4.0; i++) {
				double tmp = zr * zr - zi * zi;
				zi = 2 * zr * zi + ci;
				zr = tmp + cr;
			}
			return i;
		}

		override public int subTest (int count) {
			int i, j, c;
			double cr = -2.0, ci = 2.0;

			c = 0;
			for (j=0; j < count; j++) {
				for (i=0; i < 64; i++) {
					cr = (4.0 * (i & 63) / 64.0) - 2.0;
					ci = 2.0 - (4.0 * (i & 63) / 64.0);
					c += mandelPointTest (cr, ci);
				}
			}
			return c;
		}

		override public String name () {
			return "Mandelbrot";
		}
	}

	/*
	 *  These heap sorts use a not so common implementation of heap
	 *  sort that performs the major data swapping as a rotation 
	 *  operation.  This generally reduces data bandwidth by 50% over 
	 *  the common swap-based implementation.  It also peels off and 
	 *  simplifies the top-most case when walking down the heaps.  
	 *  Finally, it renders the inner-most comparison operation in 
	 *  such a way that it can be implemented as a conditional 
	 *  computation if the compiler and platform allow it.
	 */
	class heapSortIntTest : testCase {
		static void heapSortSiftDown (int [] arr, int top, int last) {
			int maxIdx = top;
			int tmp = arr[top];

			while (last >= (maxIdx += maxIdx)) {
				if (maxIdx != last && arr[maxIdx] < arr[maxIdx + 1])
					maxIdx++;
				if (tmp >= arr[maxIdx]) break;
				arr[top] = arr[maxIdx];
				top = maxIdx;
			}
			arr[top] = tmp;
		}

		static void heapSortSiftDown0 (int [] arr, int last) {
			int tmp;
			if (arr[1] > arr[0]) {
				tmp = arr[1];
				arr[1] = arr[0];
				heapSortSiftDown (arr, 1, last);
			} else {
				tmp = arr[0];
			}
			arr[0] = arr[last];
			arr[last] = tmp;
		}

		static void heapSort (int [] arr) {
			int qty = arr.Length;
			if (qty > 1) {
				int i = qty / 2;
				for (  qty--; i > 0; i--) heapSortSiftDown  (arr, i, qty);
				for (i = qty; i > 0; i--) heapSortSiftDown0 (arr, i);
			}
		}

		static int len = 4099;
		static int skip = 131;

		override public int subTest (int count) {
			int[] arr = new int[len];
			int i, j, k;
			for (i=j=0; i < len; i++, j += skip) {
				if (j >= len) j -= len;
				arr[i] = j;
			}
			for (k=0; k < count; k++) {
				heapSort (arr);
			}
			return arr[len-1];
		}

		override public String name () {
			return "Heapsort Integer";
		}
	}

	class heapSortFpTest : testCase {
		static void heapSortSiftDown (double [] arr, int top, int last) {
			int maxIdx = top;
			double tmp = arr[top];

			while (last >= (maxIdx += maxIdx)) {
				if (maxIdx != last && arr[maxIdx] < arr[maxIdx + 1])
					maxIdx++;
				if (tmp >= arr[maxIdx]) break;
				arr[top] = arr[maxIdx];
				top = maxIdx;
			}
			arr[top] = tmp;
		}

		static void heapSortSiftDown0 (double [] arr, int last) {
			double tmp;
			if (arr[1] > arr[0]) {
				tmp = arr[1];
				arr[1] = arr[0];
				heapSortSiftDown (arr, 1, last);
			} else {
				tmp = arr[0];
			}
			arr[0] = arr[last];
			arr[last] = tmp;
		}

		static void heapSort (double [] arr) {
			int qty = arr.Length;
			if (qty > 1) {
				int i = qty / 2;
				for (  qty--; i > 0; i--) heapSortSiftDown  (arr, i, qty);
				for (i = qty; i > 0; i--) heapSortSiftDown0 (arr, i);
			}
		}

		static int len = 4099;
		static int skip = 131;

		override public int subTest (int count) {
			double[] arr = new double[len];
			int i, j, k;
			for (i=j=0; i < len; i++, j += skip) {
				if (j >= len) j -= len;
				arr[i] = (double) j;
			}
			for (k=0; k < count; k++) {
				heapSort (arr);
			}
			return (int) arr[len-1];
		}

		override public String name () {
			return "Heapsort Double";
		}
	}

	/*
	 *  Based on the "Collatz Sequence" open math problem.  The function 
	 *  computes how many steps are required before the terminal sequence 
	 *  4, 2, 1, 4 is encountered (the open conjecture is that this 
	 *  sequence is always reached from any starting condition.)
	 */

	class collatzTest : testCase {
		static int COLLATZ_SMALL_TABLE_QTY = 16384;
		static short[] iter = new short[COLLATZ_SMALL_TABLE_QTY];
		static int ME = (int.MaxValue - 1) / 3;

		override public int subTest (int count) {
			int i, j, c = 0;
			int mv = 0;
			for (j=0; j < count; j++) {
				int mx;
				for (mx=0; mx < COLLATZ_SMALL_TABLE_QTY; mx++) {
					iter[mx] = -1;
				}

				mv = 0;
				mx = ~0;

				for (i=1; i < 113383; i++) {
					int v = i;
					int curr = 0;

					while (v > 1) {
						if (v > ME && 0 != (v & 1)) { /* Predictable. */
							Console.WriteLine ("Collatz Overflow: " + i);
							curr = -1;
							break;
						}

						if (v < COLLATZ_SMALL_TABLE_QTY) { /* Non-predicatable but clustered. */
							if (iter[v] >= 0) { /* Asymptotically predictable. */
								curr += iter[v];
								break;
							}
						}

						if (0 != (v & 1)) { /* Unpredictable. */
							v = (3*v + 1) >> 1;
						} else {
							v >>= 1;
						}

						curr ++;
					}

					if (curr > mv) { /* Highly biased. */
						mv  = curr;
						mx = i;
					}

					if (i < COLLATZ_SMALL_TABLE_QTY && curr < short.MaxValue) {
						iter[i] = (short) curr;
					}
				}

				if (mx != 106239 || mv != 223) {
					Console.WriteLine ("Unexpected result: |" + mx + "| -> " + mv);
				}

				c += mv;
			}
			return c;
		}

		override public String name () {
			return "Collatz Sequence";
		}
	}

	/*
	 *  These are the common "Stream" benchmarks for testing memory
	 *  bandwidth for floating point operations.  Basically it measures
	 *  the ability to sustain 1 or 2 streams read from memory and 1 
	 *  stream written while performing very light data dependent 
	 *  floating point operations.  Compilers have the opportunity to
	 *  detect that the arrays are non-aliasing and rerender the loops
	 *  into blocks for platforms which benefit from such 
	 *  transformations.  Alternatively, platforms/CPUs which can
	 *  handle multiple streams of simultaneous data access can remove
	 *  the need for better compilers.
	 * 
	 *  Note: these objects must be constructed with a parameter that 
	 *  guarantees that the arrays will not fit into any processor or
	 *  system cache.  The base size is 2097152.  The time results 
	 *  should be scalled accordingly.
	 */ 
	abstract class streamTest : testCase {
        static int ARRAY_SIZE_DEFAULT = 2 * 1024 * 1024; /* Default Stream Size */
		public int ARRAY_SIZE;
		public double[] a;
		public double[] b;
		public double[] c;

        int init(int sz) {
            ARRAY_SIZE = sz;
            a = new double[ARRAY_SIZE];
            b = new double[ARRAY_SIZE];
            c = new double[ARRAY_SIZE];

            for (int i = 0; i < ARRAY_SIZE; i++) {
                a[i] = 1.0;
                b[i] = 3.0;
            }
            return 0;
        }

	/*
           *  Construct the arrays to a specific size.  This will be
           *  required as processor caches increase in size over time.
           */
		public streamTest () {
            init (ARRAY_SIZE_DEFAULT);
		}

		public streamTest (int sz) {
            init (sz);
        }
	}

	class streamCopy : streamTest {
		public streamCopy (int sz) : base (sz) {
		}
        public streamCopy () : base() {
        }
		override public int subTest (int count) {
			int i, j;
			for (j=0; j < count; j++) {
				for (i=0; i < ARRAY_SIZE; i++) c[i] = a[i];
			}
			return (int) c[ARRAY_SIZE-1];
		}

		override public String name () {
			return "Stream Copy";
		}
	}

	class streamScale : streamTest {
		public streamScale (int sz) : base (sz) {
		}
        public streamScale () : base() {
        }
		override public int subTest (int count) {
			int i, j;
			for (j=0; j < count; j++) {
				for (i=0; i < ARRAY_SIZE; i++) c[i] = 3.0 * a[i];
			}
			return (int) c[ARRAY_SIZE-1];
		}

		override public String name () {
			return "Stream Scale";
		}
	}

	class streamAdd : streamTest {
		public streamAdd (int sz) : base (sz) {
		}
        public streamAdd () : base() {
        }
		override public int subTest (int count) {
			int i, j;
			for (j=0; j < count; j++) {
				for (i=0; i < ARRAY_SIZE; i++) c[i] = a[i] + b[i];
			}
			return (int) c[ARRAY_SIZE-1];
		}

		override public String name () {
			return "Stream Add";
		}
	}

	class streamTriad : streamTest {
		public streamTriad (int sz) : base (sz) {
		}
        public streamTriad () : base() {
        }
		override public int subTest (int count) {
			int i, j;
			for (j=0; j < count; j++) {
				for (i=0; i < ARRAY_SIZE; i++) c[i] = a[i] + 3.0 * b[i];
			}
			return (int) c[ARRAY_SIZE-1];
		}

		override public String name () {
			return "Stream Triad";
		}
	}

	abstract class latencyTest : testCase {
        static int ARRAY_SIZE_DEFAULT = 2 * (2 * 1024 * 1024); /* Default Stream Size */
		public int LLTEST_SIZE;
		public int[] d;

		void initArray() {
			int i;
			d = new int[LLTEST_SIZE];
			for (i=0; i < LLTEST_SIZE; i++) {
				int f = i + 43711;
				while (f >= LLTEST_SIZE) f -= LLTEST_SIZE;
				d[i] = f;
			}
		}

		public latencyTest () {
            LLTEST_SIZE = ARRAY_SIZE_DEFAULT;
			initArray ();
		}

		/*
		 *  Construct the arrays to a specific size.  This will be
		 *  required as processor caches increase in size over time.
		 */
		public latencyTest (int sz) {
			LLTEST_SIZE = 2 * sz;
			initArray ();
		}
	}

	class latencyRAM : latencyTest {
		public latencyRAM (int sz) : base (sz) {
		}
        public latencyRAM () : base() {
        }
		override public int subTest (int count) {
			int i, j, k;
			k = 0;
			for (j=0; j < count; j++) {
				for (i=0; i < 1024; i++) {
					k = d[k];
				}
			}
			return (int) k;
		}

		override public String name () {
			return "Latency RAM";
		}
	}

	class latencyL1Cache : latencyTest {
        private void init () {
			int i;
			for (i=0; i < 1024; i++) {
				d[i] &= 1023;
			}
        }
		public latencyL1Cache (int sz) : base (sz) {
            init ();
		}
        public latencyL1Cache () : base() {
            init ();
        }
		override public int subTest (int count) {
			int i, j, k;
			k = 0;
			for (j=0; j < count; j++) {
				for (i=0; i < 1024/8; i++) {
					k = d[k];
					k = d[k];
					k = d[k];
					k = d[k];
					k = d[k];
					k = d[k];
					k = d[k];
					k = d[k];
				}
			}
			return (int) k;
		}

		override public String name () {
			return "Latency L1 Cache";
		}
	}
}
