/*
 *  CritLoops benchmark by Paul Hsieh
 *
 *  Java test.
 *
 *  Copyright (c) 2012 Paul Hsieh
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *      Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *
 *      Redistributions in binary form must reproduce the above copyright notice,
 *      this list of conditions and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 *      Neither the name of CritLoops nor the names of its contributors may be used
 *      to endorse or promote products derived from this software without
 *      specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 */

import java.util.Date;
import java.util.Random;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.lang.StringBuilder;
import java.lang.Math;
import java.lang.Runtime;

public class Bench {
  static String pname = "Java";

  public static void main (String[] args) {
    double result;

    double strm;
    double strs;
    double fpuband;
    double fpulat;
    double fpulat2;
    double flow1;
    double flow2;
    double flow3;
    double memband1;
    double memband2;
    double memband3;
    double memband4;
    double memlat1;
    double memlat2;

    double strrate;
    double fpurate;
    double flowrate;
    double memrate;

    double peak_strrate;
    double peak_fpurate;
    double peak_flowrate;
    double peak_memrate;

    System.out.println ("CritLoops for Java ("+pname+")");

    if (args.length > 0) {
      pname = args[0];
    }

    System.out.println ("\tentry['" + pname + "']['sub'] = new Array (); // @CL@");

    fpuband  = runtest (new matrixMulGfx ());
    fpulat   = runtest (new mandelbrot ());
    strm     = runtest (new stringConcatTest ());
    strs     = runtest (new stringSearchTest ());
    flow1    = runtest (new heapSortIntTest ());
    flow2    = runtest (new heapSortFpTest ());
    flow3    = runtest (new collatzTest ());
    memband1 = runtest (new streamCopy ());
    memband2 = runtest (new streamScale ());
    memband3 = runtest (new streamAdd ());
    memband4 = runtest (new streamTriad ());
    memlat1  = runtest (new latencyRAM ());
    memlat2  = runtest (new latencyL1Cache ());

    /* Compute standard results */

    fpurate  = Math.pow (fpuband * fpulat, 0.5);
    strrate  = Math.pow (strm * strs, 0.5);
    flowrate = Math.pow (flow1 * flow2 * flow3, 1/3.0);
    memrate  = Math.pow (memband1 * memband2 * memband3, 1/12.0)
             * Math.pow (memband4 * memlat1 * memlat2, 1/4.0);

    /* Compute peak results */

    peak_fpurate  = fpurate;
    peak_strrate  = strrate;
    peak_flowrate = flowrate;
    peak_memrate  = memrate;

    System.out.println ("\tentry['" + pname + "']['std-rate'] = new Array(); // @CL@");
    System.out.println ("\tentry['" + pname + "']['opt-rate'] = new Array(); // @CL@");

    /* Give the results */

    System.out.println ("\tentry['" + pname + "']['std-rate'][ 'Fpu'] = " + fpurate + "; // @CL@");
    System.out.println ("\tentry['" + pname + "']['std-rate'][ 'Str'] = " + strrate + "; // @CL@");
    System.out.println ("\tentry['" + pname + "']['std-rate']['Flow'] = " + flowrate + "; // @CL@");
    System.out.println ("\tentry['" + pname + "']['std-rate'][ 'Mem'] = " + memrate + "; // @CL@");
    System.out.println ("\tentry['" + pname + "']['std'] = " + Math.pow (fpurate * strrate * flowrate * memrate, 1/4.0) + "; // @CL@");

    System.out.println ("\tentry['" + pname + "']['opt-rate'][ 'Fpu'] = " + fpurate + "; // @CL@");
    System.out.println ("\tentry['" + pname + "']['opt-rate'][ 'Str'] = " + strrate + "; // @CL@");
    System.out.println ("\tentry['" + pname + "']['opt-rate']['Flow'] = " + flowrate + "; // @CL@");
    System.out.println ("\tentry['" + pname + "']['opt-rate'][ 'Mem'] = " + memrate + "; // @CL@");
    System.out.println ("\tentry['" + pname + "']['opt'] = " + Math.pow (fpurate * strrate * flowrate * memrate, 1/4.0) + "; // @CL@");

    System.out.println ("");
  }

  public static double runtest (testCase tc) {
    long startTime = 1000;
    long stopTime = 0;
    long t0, t1, t2, t3;
    double ips;
    double DURATION = 3.0;
    int result;
    Runtime rt = Runtime.getRuntime();

    /* 
     *  Calibrate the test so that it runs for a projected 
     *  time of DURATION seconds.  This calibration step can
     *  take between 0.4s and 0.8s to complete.
     */
    int count = 1;
    while (count > 0) {
      startTime = (new Date()).getTime();
      result = tc.subTest (count);
      stopTime = (new Date()).getTime();
      if (stopTime - startTime > 0.2 * 1000.0) break;
      count = count + count;
    }

    if (0 >= count) {
      System.out.println (tc.name() + " is way too fast");
      return -1;
    }

    count = (int) (1000.0 * (double) count * DURATION / (stopTime - startTime));
    if (0 >= count) count = 1;

    /*
     *  Run the test 3 times.
     */
    rt.gc(); /* Garbage collect at the beginning */
    t0 = (new Date()).getTime();
    result = tc.subTest (count);
    t1 = (new Date()).getTime();
    result = tc.subTest (count);
    t2 = (new Date()).getTime();
    result = tc.subTest (count);
    t3 = (new Date()).getTime();

    /* The time taken for each test in clock ticks. */
    t3 -= t2;
    t2 -= t1;
    t1 -= t0;

    /* Determine median of t1, t2, t3 (throw out highest and lowest) */
    if (t3 > t2) {
      t0 = t3;
      t3 = t2;
      t2 = t0;
    }
    if (t1 >= t2) t3 = t2;
    else if (t1 >= t3) t3 = t1;

    /* 
     *  The official score of the test in units of iterations 
     *  per second.
     */
    ips = (1000.0 / t3) * count;
    /* System.out.println (tc.name() + ":\t" + ips + " iterations/second"); */
    System.out.println ("\tentry['" + pname + "']['sub']['" + tc.name() + "'] = " + ips + "; // @CL@");
    return ips;
  }
}

/* Base class for all the sub-tests */
abstract class testCase {
  /*
   *  The subTest method is supposed to implement the core 
   *  kernel that is to be tested for performance.  The count
   *  parameter is the number of times the kernel should be 
   *  run.  Each iteration of the run should be as independent
   *  as possible from previous iterations.
   */
  public abstract int subTest (int count);

  public abstract String name ();
}

/* Base class for the string tests */
abstract class stringTest extends testCase {
  static String TESTSTRING1 = "CritLoop benchmark";
  static String TESTSTRING2 = "Dot. date. datum. 123. Some more doubtful demonstration dummy data.";
}

/*
 *  Basic string concatenation performance testing.  The amount
 *  of concatenation is exaggerated versus real world applications
 *  in order to simulate the most stressful corner cases.  Various
 *  string implementation strategies, including length delimiting, 
 *  copy on write, and fully referenced based collections can have 
 *  a huge impact on the performance of string manipulation.
 */
class stringConcatTest extends stringTest {

  public int subTest (int count) {
    String a;
    int i, j, c = 0;

    a = TESTSTRING1;

    for (j=0; j < count; j++) {
      /* StringBuilder is a faster but non-thread safe stand-in for String. */
      StringBuilder accum = new StringBuilder (250 * (2 + a.length()));
      /* StringBuilder accum = new StringBuilder (); */
      for (i=0; i < 250; i++) {
        accum.append (a);      // <-- Bottleneck here.
        accum.append ("!!");
      }
      c += accum.length ();
    }
    return c;
  }

  public String name () {
    return "String Concatenation";
  }
}

/*
 *  Basic string searching performance testing.  String searching 
 *  is commonly the counter balance to the most aggressive string
 *  implementation strategies which emphasize concatenation speed.
 *  It is difficult to be completely optimal on both.  The scenarios
 *  run here also stress the algorithmic efficiency of string 
 *  searching.
 */
class stringSearchTest extends stringTest {

  public int subTest (int count) {
    String b = TESTSTRING2;
    String s = "by";
    int i, j, k, c = 0;
    /* Pattern p = Pattern.compile ("[by]"); */

    for (j=0; j < count; j++) {
      for (i=0; i < 250; i++) {
        c += b.indexOf ("ummy");
        c += b.indexOf (" data");

        /* Matcher m = p.matcher (b);
        if (m.find ()) {
          c += m.end() - 1;
        } */

        /*
		 *  Java: There does not seem to be a better way to do 
		 *  this.  Future versions of the language may include
		 *  a library function to do this, but its not clear that
		 *  it can be implemented any more efficiently.
		 */
        for (k=0; k < b.length(); k++) {
          if (s.indexOf (b.charAt(k)) > -1) {
            c += k;
            break;
          }
        }
      }
    }
    return c;
  }

  public String name () {
    return "String Scanning";
  }
}

/*
 *  These heap sorts use a not so common implementation of heap
 *  sort that performs the major data swapping as a rotation 
 *  operation.  This generally reduces data bandwidth by 50% over 
 *  the common swap-based implementation.  It also peels off and 
 *  simplifies the top-most case when walking down the heaps.  
 *  Finally, it renders the inner-most comparison operation in 
 *  such a way that it can be implemented as a conditional 
 *  computation if the compiler and platform allow it.
 */
class heapSortIntTest extends testCase {

  static void heapSortSiftDown (int [] arr, int top, int last) {
    int maxIdx = top;
    int tmp = arr[top];

    while (last >= (maxIdx += maxIdx)) {
      if (maxIdx != last && arr[maxIdx] < arr[maxIdx + 1])
        maxIdx++;
      if (tmp >= arr[maxIdx]) break;
      arr[top] = arr[maxIdx];
      top = maxIdx;
    }
    arr[top] = tmp;
  }

  static void heapSortSiftDown0 (int [] arr, int last) {
    int tmp;
    if (arr[1] > arr[0]) {
      tmp = arr[1];
      arr[1] = arr[0];
      heapSortSiftDown (arr, 1, last);
    } else {
      tmp = arr[0];
    }
    arr[0] = arr[last];
    arr[last] = tmp;
  }

  static void heapSort (int [] arr) {
    int qty = arr.length;
    if (qty > 1) {
      int i = qty / 2;
      for (  qty--; i > 0; i--) heapSortSiftDown  (arr, i, qty);
      for (i = qty; i > 0; i--) heapSortSiftDown0 (arr, i);
    }
  }

  static int len = 4099;
  static int skip = 131;

  public int subTest (int count) {
    int[] arr = new int[len];
    int i, j, k;
    for (i=j=0; i < len; i++, j += skip) {
      if (j >= len) j -= len;
      arr[i] = j;
    }
    for (k=0; k < count; k++) {
      heapSort (arr);
    }
    return arr[len-1];
  }

  public String name () {
    return "Heapsort Integer";
  }
}

class heapSortFpTest extends testCase {

  static void heapSortSiftDown (double [] arr, int top, int last) {
    int maxIdx = top;
    double tmp = arr[top];

    while (last >= (maxIdx += maxIdx)) {
      if (maxIdx != last && arr[maxIdx] < arr[maxIdx + 1])
        maxIdx++;
      if (tmp >= arr[maxIdx]) break;
      arr[top] = arr[maxIdx];
      top = maxIdx;
    }
    arr[top] = tmp;
  }

  static void heapSortSiftDown0 (double [] arr, int last) {
    double tmp;
    if (arr[1] > arr[0]) {
      tmp = arr[1];
      arr[1] = arr[0];
      heapSortSiftDown (arr, 1, last);
    } else {
      tmp = arr[0];
    }
    arr[0] = arr[last];
    arr[last] = tmp;
  }

  static void heapSort (double [] arr) {
    int qty = arr.length;
    if (qty > 1) {
      int i = qty / 2;
      for (  qty--; i > 0; i--) heapSortSiftDown  (arr, i, qty);
      for (i = qty; i > 0; i--) heapSortSiftDown0 (arr, i);
    }
  }

  static int len = 4099;
  static int skip = 131;

  public int subTest (int count) {
    double[] arr = new double[len];
    int i, j, k;
    for (i=j=0; i < len; i++, j += skip) {
      if (j >= len) j -= len;
      arr[i] = (double) j;
    }
    for (k=0; k < count; k++) {
      heapSort (arr);
    }
    return (int) arr[len-1];
  }

  public String name () {
    return "Heapsort Double";
  }
}

/*
 *  Based on the "Collatz Sequence" open math problem.  The function 
 *  computes how many steps are required before the terminal sequence 
 *  4, 2, 1, 4 is encountered (the open conjecture is that this 
 *  sequence is always reached from any starting condition.)
 */

class collatzTest extends testCase {
  static int COLLATZ_SMALL_TABLE_QTY = 16384;
  static short[] iter = new short[COLLATZ_SMALL_TABLE_QTY];
  static int ME = (Integer.MAX_VALUE - 1) / 3;

  public int subTest (int count) {
    int i, j, c = 0;
    int mv = 0;

    for (j=0; j < count; j++) {
      int mx;

      for (mx=0; mx < COLLATZ_SMALL_TABLE_QTY; mx++) {
        iter[mx] = -1;
      }

      mv = 0;
      mx = ~0;

      for (i=1; i < 113383; i++) {
        int v = i;
        int curr = 0;

        while (v > 1) {
          if (v > ME && 0 != (v & 1)) { /* Predictable. */
            System.out.println ("Collatz Overflow: " + i);
            curr = -1;
            break;
          }

          if (v < COLLATZ_SMALL_TABLE_QTY) { /* Non-predicatable but clustered. */
            if (iter[v] >= 0) { /* Asymptotically predictable. */
              curr += iter[v];
              break;
            }
          }

          if (0 != (v & 1)) { /* Unpredictable. */
            v = (3*v + 1) >> 1;
          } else {
            v >>= 1;
          }

          curr ++;
        }

        if (curr > mv) { /* Highly biased. */
          mv  = curr;
          mx = i;
        }

        if (i < COLLATZ_SMALL_TABLE_QTY && curr < Short.MAX_VALUE) {
          iter[i] = (short) curr;
        }
      }

      if (mx != 106239 || mv != 223) {
        System.out.println ("Unexpected result: |" + mx + "| -> " + mv);
      }

      c += mv;
    }

    return c;
  }

  public String name () {
    return "Collatz Sequence";
  }
}

/*
 *  These are the common "Stream" benchmarks for testing memory
 *  bandwidth for floating point operations.  Basically it measures
 *  the ability to sustain 1 or 2 streams read from memory and 1 
 *  stream written while performing very light data dependent 
 *  floating point operations.  Compilers have the opportunity to
 *  detect that the arrays are non-aliasing and rerender the loops
 *  into blocks for platforms which benefit from such 
 *  transformations.  Alternatively, platforms/CPUs which can
 *  handle multiple streams of simultaneous data access can remove
 *  the need for better compilers.
 * 
 *  Note: these objects must be constructed with a parameter that 
 *  guarantees that the arrays will not fit into any processor or
 *  system cache.  The base size is 2097152.  The time results 
 *  should be scalled accordingly.
 */ 
abstract class streamTest extends testCase {
  int ARRAY_SIZE;
  static int ARRAY_SIZE_DEFAULT = 2 * 1024 * 1024;
  double[] a;
  double[] b;
  double[] c;

  int initArrays () {
    int i;
    int ret = 0;

    try {
      a = new double[ARRAY_SIZE];
      b = new double[ARRAY_SIZE];
      c = new double[ARRAY_SIZE];
    }
    catch (java.lang.OutOfMemoryError e) {
      a = null;
      b = null;
      c = null;
      if (ARRAY_SIZE < 32*1024) throw e;
      return -1;
    }

    for (i=0; i < ARRAY_SIZE; i++) {
      a[i] = 1.0;
      b[i] = 3.0;
    }
    return 0;
  }

  int generalInit (int sz) {
    byte[] slack = new byte[128*1024]; /* Reserved memory for after init */

    ARRAY_SIZE = sz;
    while (initArrays () < 0) {
      sz = sz / 2;
      ARRAY_SIZE = sz;
    }

    slack = null;
    return ARRAY_SIZE;
  }

  public streamTest () {
    int su = generalInit (64 * ARRAY_SIZE_DEFAULT);
  }

  public streamTest (int sz) {
    int su = generalInit (sz);
  }
}

class streamCopy extends streamTest {
  public streamCopy (int sz) {
    super (sz);
  }
  public streamCopy () {
    super ();
  }
  public int subTest (int count) {
    int i, j;
    count = (ARRAY_SIZE_DEFAULT * count + ARRAY_SIZE - 1) / ARRAY_SIZE;
    for (j=0; j < count; j++) {
      for (i=0; i < ARRAY_SIZE; i++) c[i] = a[i];
    }
    return (int) c[ARRAY_SIZE-1];
  }

  public String name () {
    return "Stream Copy";
  }
}

class streamScale extends streamTest {
  public streamScale (int sz) {
    super (sz);
  }
  public streamScale () {
    super ();
  }
  public int subTest (int count) {
    int i, j;
    count = (ARRAY_SIZE_DEFAULT * count + ARRAY_SIZE - 1) / ARRAY_SIZE;
    for (j=0; j < count; j++) {
      for (i=0; i < ARRAY_SIZE; i++) c[i] = 3.0 * a[i];
    }
    return (int) c[ARRAY_SIZE-1];
  }

  public String name () {
    return "Stream Scale";
  }
}

class streamAdd extends streamTest {
  public streamAdd (int sz) {
    super (sz);
  }
  public streamAdd () {
    super ();
  }
  public int subTest (int count) {
    int i, j;
    count = (ARRAY_SIZE_DEFAULT * count + ARRAY_SIZE - 1) / ARRAY_SIZE;
    for (j=0; j < count; j++) {
      for (i=0; i < ARRAY_SIZE; i++) c[i] = a[i] + b[i];
    }
    return (int) c[ARRAY_SIZE-1];
  }

  public String name () {
    return "Stream Add";
  }
}

class streamTriad extends streamTest {
  public streamTriad (int sz) {
    super (sz);
  }
  public streamTriad () {
    super ();
  }
  public int subTest (int count) {
    int i, j;
    count = (ARRAY_SIZE_DEFAULT * count + ARRAY_SIZE - 1) / ARRAY_SIZE;
    for (j=0; j < count; j++) {
      for (i=0; i < ARRAY_SIZE; i++) c[i] = a[i] + 3.0 * b[i];
    }
    return (int) c[ARRAY_SIZE-1];
  }

  public String name () {
    return "Stream Triad";
  }
}

abstract class latencyTest extends testCase {
  int LLTEST_SIZE;
  int[] d;

  void initArrays () {
    int i;
    for (i=0; i < LLTEST_SIZE; i++) {
      int f = i + 43711;
      while (f >= LLTEST_SIZE) f -= LLTEST_SIZE;
      d[i] = f;
    }
  }

  public latencyTest () {
    LLTEST_SIZE = 2 * 2 * 1024 * 1024;
    d = new int[LLTEST_SIZE];
    initArrays ();
  }

  public latencyTest (int sz) {
    LLTEST_SIZE = 2 * sz;
    d = new int[LLTEST_SIZE];
    initArrays ();
  }
}

/* Perform a linked list traversal test in a large section of memory. */
class latencyRAM extends latencyTest {
  public latencyRAM (int sz) {
    super (sz);
  }
  public latencyRAM () {
    super ();
  }
  public int subTest (int count) {
    int i, j, k;
    k = 0;
    for (j=0; j < count; j++) {
      for (i=0; i < 1024; i++) {
        k = d[k];
      }
    }
    return (int) k;
  }
  public String name () {
    return "Latency RAM";
  }
}

/* Perform a linked list traversal test in a very small section of memory. */
class latencyL1Cache extends latencyTest {
  void init () {
    int i;
    for (i=0; i < 1024; i++) {
      d[i] &= 1023;
    }
  }
  public latencyL1Cache (int sz) {
    super (sz);
    init ();
  }
  public latencyL1Cache () {
    super ();
    init ();
  }
  public int subTest (int count) {
    int i, j, k;
    k = 0;
    for (j=0; j < count; j++) {
      for (i=0; i < 1024/8; i++) {
        k = d[k];
        k = d[k];
        k = d[k];
        k = d[k];
        k = d[k];
        k = d[k];
        k = d[k];
        k = d[k];
      }
    }
    return (int) k;
  }
  public String name () {
    return "Latency L1 Cache";
  }
}

/*
 *  A standard 4x4 matrix multiplication.  Although matrix by vector
 *  is more common, it can be considered a subset of this operation 
 *  (the performance characteristics should be the same).  One matrix 
 *  is assumed (runtime) constant and pre-transposed for better memory 
 *  access.  The most inner loop is unrolled, as it is assumed that 
 *  developers interested in this kernel would be concerned with
 *  performance from the start.  The primary platform potential here 
 *  should be to leverage the SIMD capabilities of modern processors.
 */
class matrixMulGfx extends testCase {
  public int subTest (int count) {
    float[][] preTransposed = new float[][] {
      { 0, 1, 0, 0 },
      { 1, 0, 0, 0 },
      { 0, 0, 1, 0 },
      { 0, 0, 0, 1 }
    };
    float[][] source = new float[4][4];
    float[][] target = new float[4][4];
    int i, j;
    int k;
    java.util.Random prng = new java.util.Random (System.currentTimeMillis ());

    for (i=0; i < 4; i++) {
      for (j=0; j < 4; j++) {
        source[i][j] = prng.nextFloat ();
      }
    }

    for (k=0; k < count; k++) {
      for (j=0; j < 1000; j++) {
        for (i=0; i < 4; i++) {
          /* Unrolled -- as opposed to FPUMark which was a loop */
          target[i][0] = preTransposed[0][0] * source[i][0]
                       + preTransposed[0][1] * source[i][1]
                       + preTransposed[0][2] * source[i][2]
                       + preTransposed[0][3] * source[i][3];
          target[i][1] = preTransposed[1][0] * source[i][0]
                       + preTransposed[1][1] * source[i][1]
                       + preTransposed[1][2] * source[i][2]
                       + preTransposed[1][3] * source[i][3];
          target[i][2] = preTransposed[2][0] * source[i][0]
                       + preTransposed[2][1] * source[i][1]
                       + preTransposed[2][2] * source[i][2]
                       + preTransposed[2][3] * source[i][3];
          target[i][3] = preTransposed[3][0] * source[i][0]
                       + preTransposed[3][1] * source[i][1]
                       + preTransposed[3][2] * source[i][2]
                       + preTransposed[3][3] * source[i][3];
        }
        for (i=0; i < 4; i++) {
          source[i][0] = target[i][0];
          source[i][1] = target[i][1];
          source[i][2] = target[i][2];
          source[i][3] = target[i][3];
        }
      }
    }
    return (int) (100.0 * target[3][3]);
  }

  public String name () {
    return "Matrix Multiply GFX";
  }
}

class mandelbrot extends testCase {
  int MAX_ITERS = 2048;

  public int mandelPointTest (double cr, double ci) {
    double zr, zi;
    int i;

    zr = cr; zi = ci;

    for (i=0; i < MAX_ITERS && zr*zr + zi*zi < 4.0; i++) {
      double tmp = zr * zr - zi * zi;
      zi = 2 * zr * zi + ci;
      zr = tmp + cr;
    }
    return i;
  }

  public int subTest (int count) {
    int i, j, c;
    double cr = -2.0, ci = 2.0;

    c = 0;
    for (j=0; j < count; j++) {
      for (i=0; i < 64; i++) {
        cr = (4.0 * (i & 63) / 64.0) - 2.0;
        ci = 2.0 - (4.0 * (i & 63) / 64.0);
        c += mandelPointTest (cr, ci);
      }
    }
    return c;
  }

  public String name () {
    return "Mandelbrot";
  }
}

